
class Observations(object):
    def __init__(self, slices=None, times=None):
        self.slices = []
        self.times = times
        self._sequence = []
        if slices:
            for slic in slices:
                self.append(slic)
                for elem in slic:
                    self._sequence.append(elem)

    def append(self, l):
        try:
            _ = iter(l)
        except TypeError:
            print(l, 'is not iterable')
        for pc in l:
            if pc not in list(range(12)):
                raise TypeError(pc, "is not suported")
        self.slices.append(l)

    def __iter__(self):
        for elem in self._sequence:
            yield elem

    def __getitem__(self, key):
        return self._sequence[key]

    def __len__(self):
        return len(self._sequence)

    def slice_indexes(self):
        for idx, slic in enumerate(self.slices):
            for _ in slic:
                yield idx

    def tuples(self):
        for idx, slic in enumerate(self.slices):
            for elem in slic:
                yield elem, idx

    @staticmethod
    def from_tuples(l):
        slices = Observations()
        max_index = max(l, key=lambda x: x[1])[1]
        [slices.append([]) for _ in range(max_index + 1)]
        for elem, idx in l:
            slices.slices[idx].append(elem)
        return slices



import justkeydding.parameters


def get_key_from_filename(filename):
    """Returns the key of a midi file if it is a postfix of the filename"""
    key = filename[:-4].split('_')[-1]
    keys_plus_enharmonics = list(justkeydding.parameters.states) + list(justkeydding.parameters.enharmonics.values())
    return key if key in keys_plus_enharmonics else 'x'


def is_key_guess_correct(ground_truth, guess):
    """Returns whether a key guess is correct or not"""
    if ground_truth in justkeydding.parameters.states:
        iscorrect = True if ground_truth == guess else False
    elif ground_truth in list(justkeydding.parameters.enharmonics.values()):
        if guess in justkeydding.parameters.enharmonics and ground_truth == justkeydding.parameters.enharmonics[guess]:
            iscorrect = True
        else:
            iscorrect = False
    return iscorrect



from justkeydding.parameters import key_profiles, key_transitions
import random
import logging


class Generator:
    def __init__(self, kp_max_range, kt_max_range):
        self.logger = logging.getLogger('generator')
        self.kp_maximum_range = kp_max_range
        self.kt_maximum_range = kt_max_range
        self.kp_id = 0
        self.kt_id = 0
        self.logger.info("Generator() <- kp_max_range={}, kt_max_range={}".format(self.kp_maximum_range, self.kt_maximum_range))

    def generate_key_profile(self):
        ''' Generate a set of random key profiles '''
        self.logger.info('Start generate_key_profile()')
        major = self.generate_major_key_profile()
        minor = self.generate_minor_key_profile()
        key_profile = major + minor
        key_profile_name = 'kp{}'.format(self.kp_id)
        self.kp_id += 1
        key_profiles.insert_new(key_profile_name, key_profile)
        self.logger.info('Done generate_key_profile() -> key_profile_name={}'.format(key_profile_name))
        return key_profile_name

    def generate_major_key_profile(self):
        self.logger.info('Start generate_major_key_profile()')
        key_profile = [
            0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
        ]
        tonic_dominant = [0, 7]
        diatonic = [2, 4, 5, 9, 11]
        non_diatonic = [1, 3, 6, 8, 10]
        groups = [
            (tonic_dominant, 0),
            (diatonic, 1),
            (non_diatonic, 2)
        ]
        for group, hierarchy in groups:
            for idx in group:
                r = random.randrange(int(self.kp_maximum_range / 2**hierarchy))
                key_profile[idx] = r
                self.logger.debug('{}: {}'.format(idx, r))
        s = sum(key_profile)
        key_profile = [x/s for x in key_profile]
        self.logger.info('Done generate_major_key_profile() -> key_profile={}'.format(key_profile))
        return key_profile

    def generate_minor_key_profile(self):
        self.logger.info('Start generate_minor_key_profile()')
        key_profile = [
            0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
        ]
        tonic_dominant = [0, 7]
        harmonic = [2, 3, 5, 8, 11]
        natural_melodic = [9, 10]
        non_diatonic = [1, 4, 6]
        groups = [
            (tonic_dominant, 0),
            (harmonic, 1),
            (natural_melodic, 2),
            (non_diatonic, 3)
        ]
        for group, hierarchy in groups:
            for idx in group:
                r = random.randrange(int(self.kp_maximum_range / 2**hierarchy))
                key_profile[idx] = r
                self.logger.debug('{}: {}'.format(idx, r))
        s = sum(key_profile)
        key_profile = [x/s for x in key_profile]
        self.logger.info('Done generate_minor_key_profile() -> key_profile={}'.format(key_profile))
        return key_profile

    def generate_key_transition(self):
        ''' Generate a set of random key transitions '''
        self.logger.info("Start generate_key_transition()")
        kt_type = random.choice(['geometric', 'geometric', 'geometric', 'geometric', 'geometric', 'geometric', 'geometric', 'geometric', 'geometric', 'stochastic'])
        if kt_type == 'geometric':
            ratio = random.random() * 48.0 + 2.0
            key_transition = self.generate_geometric_key_transition(ratio)
            key_transition_name = 'ktg{}'.format(self.kt_id)
            key_transitions.store_ratio(key_transition_name, ratio)
        elif kt_type == 'stochastic':
            key_transition = self.generate_stochastic_key_transition()
            key_transition_name = 'kts{}'.format(self.kt_id)
        self.kt_id += 1
        key_transitions.insert_new(key_transition_name, key_transition)
        self.logger.info('Done generate_key_transition() -> key_transition_name={}'.format(key_transition_name))
        return key_transition_name

    def generate_geometric_key_transition(self, ratio):
        self.logger.info('Start generate_geometric_key_transition() <- ratio={}'.format(ratio))
        key_transition = [
            0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
            0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
        ]
        groups = [
            [0],
            [5, 7, 12, 21],
            [14, 16, 17, 19],
            [2, 3, 9, 10],
            [4, 8, 22, 23],
            [1, 11],
            [15, 18],
            [13, 20],
            [6]
        ]
        number_of_groups = len(groups) - 1
        for group_id, keys in enumerate(groups):
            for key in keys:
                r = ratio**(number_of_groups - group_id)
                key_transition[key] = r
                self.logger.debug('{}: {}'.format(key, r))
        s = sum(key_transition)
        key_transition = [x/s for x in key_transition]
        self.logger.info('Done generate_geometric_key_transition() -> key_transition={}'.format(key_transition))
        return key_transition

    def generate_stochastic_key_transition(self):
        self.logger.info('Start generate_stochastic_key_transition()')
        key_transition = [
            0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
            0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
        ]
        groups = [
            [0],
            [5, 7, 12, 21],
            [14, 16, 17, 19],
            [2, 3, 9, 10],
            [4, 8, 22, 23],
            [1, 11],
            [15, 18],
            [13, 20],
            [6]
        ]
        for group_id, keys in enumerate(groups):
            for key in keys:
                r = random.randrange(int(self.kt_maximum_range / 2**group_id))
                key_transition[key] = r
                self.logger.debug('{}: {}'.format(key, r))
        s = sum(key_transition)
        key_transition = [x/s for x in key_transition]
        self.logger.info('Done generate_stochastic_key_transition() -> key_transition={}'.format(key_transition))
        return key_transition

    def generate_key_profiles(self, count):
        ''' Generate <count> key_profiles '''
        self.logger.info('Start generate_key_profiles() <- count={}'.format(count))
        key_profiles = [self.generate_key_profile() for _ in range(count)]
        self.logger.info('Done generate_key_profiles() -> key_profiles={}'.format(key_profiles))
        return key_profiles

    def generate_key_transitions(self, count):
        ''' Generate <count> key_transitions '''
        self.logger.info('Start generate_key_transitions() <- count={}'.format(count))
        key_transitions = [self.generate_key_transition() for _ in range(count)]
        self.logger.info('Done generate_key_transitions() -> key_transitions={}'.format(key_transitions))
        return key_transitions



from justkeydding.parameters import key_profiles, key_transitions
import subprocess
import logging
import itertools
from multiprocessing.dummy import Pool as ThreadPool

class Ensembler:
    def __init__(self, profiles, transitions):
        self.logger = logging.getLogger('ensembler')
        self.logger.info('Ensembler() <- profiles={}, transitions={}'.format(profiles, transitions))
        self.profiles = profiles
        self.transitions = transitions

    def grade_key_profiles(self, key_profiles, key_transition_name):
        ''' Grade a list of key profiles '''
        self.logger.info('Start grade_key_profiles() <- key_profiles={}, key_transition_name={}'.format(key_profiles, key_transition_name))
        grading = [self.evaluate(x, key_transition_name) for x in key_profiles]
        grading = sorted(grading, key=lambda score: score[0])
        self.logger.info('Done grade_key_profiles() -> grading={}'.format(grading))
        return grading

    def grade_key_transitions(self, key_profile_name, key_transitions):
        ''' Grade a list of key transitions '''
        self.logger.info('Start grade_key_transitions() <- key_profile_name={}, key_transitions ={}'.format(key_profile_name, key_transitions))
        grading = [self.evaluate(key_profile_name, x) for x in key_transitions]
        grading = sorted(grading, key=lambda score: score[0])
        self.logger.info('Done grade_key_transitions() -> grading={}'.format(grading))
        return grading

    def get_ensemble(self, mixed_profiles=False):
        ''' Create the ensemble '''
        self.logger.info('Start get_ensemble() <- mixedProfiles={}'.format(mixed_profiles))
        if mixed_profiles:
            profiles = [key_profiles.mix(m[0], m[1]) for m in itertools.product(self.profiles, self.profiles)]
        else:
            profiles = self.profiles
        self.ensemble = [(e[0], e[1]) for e in itertools.product(profiles, self.transitions)]
        return self.ensemble

    def evaluate(self, filename, mixed_profiles=False):
        ''' Evaluate a key profile '''
        self.logger.info('Start evaluate() <- filename={}'.format(filename))
        if not hasattr(self, 'ensemble'):
            self.get_ensemble(mixed_profiles)
        with ThreadPool(4) as p:
            features = p.map(lambda e: self.run_keydetection(filename, e), self.ensemble)
        self.logger.info('Done evaluate() -> features={}'.format(features))
        return features

    def run_keydetection(self, filename, ensemble):
        self.logger.debug('run_keydetection() <- filename={}, ensemble={}'.format(filename, ensemble))
        key_profile_name = ensemble[0]
        key_transition_name = ensemble[1]
        kp_string = key_profiles.get_as_string(key_profile_name)
        kt_string = key_transitions.get_as_string(key_transition_name)
        justkeydding = subprocess.Popen(
                ('bin/justkeydding',
                '-p',
                '-K',
                '{}'.format(kp_string),
                '-T',
                '{}'.format(kt_string),
                filename),
                stdout=subprocess.PIPE)
        output, _ = justkeydding.communicate()
        features = output.split()
        if len(features) != 24:
            self.logger.error('Failed while running justkeydding, output: {}'.format(output))
            features = ['-1000000'] * 24
        features = [float(x) for x in features]
        self.logger.debug('{} {}: {}'.format(filename, ensemble, features))
        return features



from justkeydding.parameters import key_profiles, key_transitions
import subprocess
import logging
from multiprocessing.dummy import Pool
import threading

class Evaluator:
    def __init__(self, dataset):
        with open(dataset) as f:
            self.logger = logging.getLogger('evaluator')
            self.logger.info('Evaluator() <- dataset={}'.format(dataset))
            self.files = f.readlines()
            self.files = [x.strip() for x in self.files]
            self.logger.info('Evaluator() -> self.files={}'.format(self.files))

    def grade_key_profiles(self, key_profiles, key_transition_name):
        ''' Grade a list of key profiles '''
        self.logger.info('Start grade_key_profiles() <- key_profiles={}, key_transition_name={}'.format(key_profiles, key_transition_name))
        grading = [self.evaluate(x, key_transition_name) for x in key_profiles]
        grading = sorted(grading, key=lambda score: score[0])
        self.logger.info('Done grade_key_profiles() -> grading={}'.format(grading))
        return grading

    def grade_key_transitions(self, key_profile_name, key_transitions):
        ''' Grade a list of key transitions '''
        self.logger.info('Start grade_key_transitions() <- key_profile_name={}, key_transitions ={}'.format(key_profile_name, key_transitions))
        grading = [self.evaluate(key_profile_name, x) for x in key_transitions]
        grading = sorted(grading, key=lambda score: score[0])
        self.logger.info('Done grade_key_transitions() -> grading={}'.format(grading))
        return grading

    def evaluate(self, key_profile_name, key_transition_name):
        ''' Evaluate a key profile '''
        self.logger.info('Start evaluate() <- key_profile_name={}, key_transition_name={}'.format(key_profile_name, key_transition_name))
        # error_list = [self.run_keydetection(filename, key_profile_name, key_transition_name)
        #              for filename in self.files]
        with Pool(4) as p:
            error_list = p.map(lambda f: self.run_keydetection(f, key_profile_name, key_transition_name), self.files)
        total_error = sum(error_list)
        self.logger.info('Done evaluate() -> total_error={}'.format(total_error))
        return (total_error, key_profile_name, key_transition_name)

    def run_keydetection(self, filename, key_profile_name, key_transition_name):
        kp_string = key_profiles.get_as_string(key_profile_name)
        kt_string = key_transitions.get_as_string(key_transition_name)
        # self.logger.debug('kp_string:"{}", kt_string:"{}"'.format(kp_string, kt_string))
        justkeydding = subprocess.Popen(
                ('bin/justkeydding',
                '-e',
                '-K',
                '{}'.format(kp_string),
                '-T',
                '{}'.format(kt_string),
                filename),
                stdout=subprocess.PIPE)
        output, _ = justkeydding.communicate()
        try:
            score = float(output)
        except ValueError:
            self.logger.error('Failed while running justkeydding on file {}, output: {}'.format(filename, output))
            score = 0.0
        error = 1 - score**2
        self.logger.debug('{}: {}'.format(filename, error))
        return error




from justkeydding.optimizer import ensembler
import collections
import os
import logging
import logging.handlers
import logging.config
import numpy as np
import sys

logging_dict = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'date': {
            'format': '%(asctime)s [%(name)-15s] %(levelname)-7s: %(message)s',
        },
        'no_date': {
            'format': '[%(name)-15s] %(levelname)-7s: %(message)s',
        },
    },
    'handlers': {
        'file_handler': {
            'class': 'logging.handlers.TimedRotatingFileHandler',
            'filename': 'logs/optimizer.log',
            'when': 'H',
            'interval': 1,
            'backupCount': 360,
            'formatter': 'date',
            'encoding': 'utf8'
        },
        'console': {
            'class': 'logging.StreamHandler',
            'level': 'DEBUG',
            'formatter': 'no_date'
        }
    },
    'loggers': {
        '': {
            'handlers': ['file_handler', 'console'],
            'level': 'DEBUG',
            'propagate': True
        }
    }
}


class DatasetCreator():
    keys_single_letter = {
        'C': 0, 'C#': 1, 'Db': 1, 'D': 2,
        'D#': 3, 'Eb': 3, 'E': 4, 'F': 5,
        'F#': 6, 'Gb': 6, 'G': 7, 'G#': 8,
        'Ab': 8, 'A': 9, 'A#': 10, 'Bb': 10,
        'B': 11,

        'c': 12, 'c#': 13, 'db': 13, 'd': 14,
        'd#': 15, 'eb': 15, 'e': 16, 'f': 17,
        'f#': 18, 'gb': 18, 'g': 19, 'g#': 20,
        'ab': 20, 'a': 21, 'a#': 22, 'bb': 22,
        'b': 23
    }
    keys_duple = {
        ('c', 'major'): 0, ('c#', 'major'): 1,
        ('db', 'major'): 1, ('d', 'major'): 2,
        ('d#', 'major'): 3, ('eb', 'major'): 3,
        ('e', 'major'): 4, ('f', 'major'): 5,
        ('f#', 'major'): 6, ('gb', 'major'): 6,
        ('g', 'major'): 7, ('g#', 'major'): 8,
        ('ab', 'major'): 8, ('a', 'major'): 9,
        ('a#', 'major'): 10, ('bb', 'major'): 10,
        ('b', 'major'): 11,

        ('c', 'minor'): 12, ('c#', 'minor'): 13,
        ('db', 'minor'): 13, ('d', 'minor'): 14,
        ('d#', 'minor'): 15, ('eb', 'minor'): 15,
        ('e', 'minor'): 16, ('f', 'minor'): 17,
        ('f#', 'minor'): 18, ('gb', 'minor'): 18,
        ('g', 'minor'): 19, ('g#', 'minor'): 20,
        ('ab', 'minor'): 20, ('a', 'minor'): 21,
        ('a#', 'minor'): 22, ('bb', 'minor'): 22,
        ('b', 'minor'): 23
    }

    def __init__(self, dataset_folder):
        self.logger = logging.getLogger('DatasetCreator')
        self.logger.info(
            'DatasetCreator() <- dataset_folder={}'.format(dataset_folder))
        self.dataset_okay = False
        self.has_features = False
        self.ensemble = None
        # Check critical stuff:
        # dataset dir
        if not os.path.isdir(dataset_folder):
            self.logger.error('The dataset provided is not a folder')
            exit()
        if dataset_folder.endswith('/'):
            self.dataset_folder = dataset_folder[:-1]
        else:
            self.dataset_folder = dataset_folder
        if not any(c.isalpha() for c in self.dataset_folder):
            self.logger.warning('The dataset folder, {}, does not contain any parsable name.'.format(self.dataset_folder))
            self.name = 'unknown_dataset'

        self.name = self.dataset_folder.rsplit('/')[-1]
        self.logger.info('Dataset name: {}'.format(self.name))
        self.feature_folder = os.path.join(dataset_folder, 'features')
        # features folder
        if not os.path.isdir(self.feature_folder):
            self.logger.error(('The dataset folder must '
                               'contain a "features" folder'))
            exit()
        self.files = []
        self.files_no_extension = []
        # files there and right formats
        for f in os.listdir(self.feature_folder):
            tokens = f.rsplit('.', 1)
            if tokens[1] not in ['mid', 'csv', 'wav']:
                self.logger.error('Unknown file type {}. Expecting .wav, .csv, or .mid'.format(f))
                exit()
            no_extension = tokens[0]
            self.files.append(f)
            self.files_no_extension.append(no_extension)
        if not self.files:
            self.logger.error('The features folder seems to be empty.')
            exit()
        # Now, not so critical stuff
        self.annotation_folder = os.path.join(dataset_folder, 'annotations')
        if not os.path.isdir(self.annotation_folder):
            self.logger.warning('The dataset folder does not contain annotations, assuming this was expected')
            self.annotation_folder = None
        # And not critical at all
        self.splits_folder = os.path.join(dataset_folder, 'splits')
        if not os.path.isdir(self.splits_folder):
            self.logger.info('No data splits provided. Working on the entire dataset')
            self.splits_folder = None
        self.features = collections.OrderedDict()
        self.annotations = collections.OrderedDict()
        self.splits = {}
        if self.splits_folder:
            for f in os.listdir(self.splits_folder):
                tokens = f.rsplit('.', 1)
                filepath = os.path.join(self.splits_folder, f)
                if len(tokens) < 2 or tokens[1] not in ['txt']:
                    self.logger.warning('Expected {} to be a .txt file.'.format(filepath))
                no_extension = tokens[0]
                with open(filepath, newline='') as fd:
                    split = fd.readlines()
                    split = [x.strip() for x in split]
                    self.splits[no_extension] = split
                self.logger.info('Found the following splits: {}'.format(self.splits))
        # If we are here, everything should be okay
        self.dataset_okay = True

    def parse_label(self, annotation):
        annotation = annotation.strip().split()
        label = -1
        if len(annotation) == 1:
            key = annotation[0]
            if key in DatasetCreator.keys_single_letter:
                label = DatasetCreator.keys_single_letter[key]
        elif len(annotation) == 2:
            key = (annotation[0].lower(), annotation[1].lower())
            if key in DatasetCreator.keys_duple:
                label = DatasetCreator.keys_duple[key]
        if label == -1:
            self.logger.error('Did not recognize the format of label {}'.format(annotation))
        return label

    def compute_features(self, key_profiles, key_transitions, mixed_profiles=False):
        self.logger.info('compute_features() <- key_profiles={}, key_transitions={}, mixed_profiles={}'.format(key_profiles, key_transitions, mixed_profiles))
        if not self.dataset_okay:
            return
        self.key_profiles = key_profiles
        self.key_transitions = key_transitions
        self.mixed_profiles = mixed_profiles
        ens = ensembler.Ensembler(key_profiles, key_transitions)
        for i in range(len(self.files)):
            feature_filepath = os.path.join(self.feature_folder, self.files[i])
            features = ens.evaluate(feature_filepath, mixed_profiles)
            features = [f for l in features for f in l]
            feature_array = np.array(features)
            self.features[self.files_no_extension[i]] = feature_array
            if self.annotation_folder:
                annotation_name = '{}.key'.format(self.files_no_extension[i])
                annotation_filepath = os.path.join(self.annotation_folder, annotation_name)
                if not os.path.exists(annotation_filepath):
                    self.logger.error('Could not find file {}'.format(annotation_filepath))
                    return
                with open(annotation_filepath, newline='') as fd:
                    annotation_data = fd.readlines()[0]
                label = self.parse_label(annotation_data)
                self.annotations[self.files_no_extension[i]] = label
        self.has_features = True

    def write(self, output_dir='./datasets'):
        self.logger.info('write() <- output_dir={}, dataset_name={}'.format(output_dir, self.name))
        if not self.has_features:
            return
        if not os.path.exists(output_dir):
            self.logger.warning('Output folder {} does not exist. Trying to create it'.format(output_dir))
            os.makedirs(output_dir)
            self.logger.warning('Success.')
        ensemble_filename = '{}_ensemble.txt'.format(self.name)
        ensemble_filepath = os.path.join(output_dir, ensemble_filename)
        self.logger.info('writing {}'.format(ensemble_filepath))
        with open(ensemble_filepath, 'w') as fd:
            fd.write('{}\n'.format(', '.join(self.key_profiles)))
            fd.write('{}\n'.format(', '.join(self.key_transitions)))
            fd.write('mixed_profiles={}\n'.format(self.mixed_profiles))
        features_filename = '{}_features.pkl'.format(self.name)
        feature_filepath = os.path.join(output_dir, features_filename)
        feature_array = list(self.features.values())
        self.logger.info('writing {}'.format(feature_filepath))
        np.array(feature_array).dump(feature_filepath)
        if self.annotations:
            annotation_filename = '{}_annotations.pkl'.format(self.name)
            annotation_filepath = os.path.join(output_dir, annotation_filename)
            annotation_array = list(self.annotations.values())
            self.logger.info('writing {}'.format(annotation_filepath))
            np.array(annotation_array).dump(annotation_filepath)
        for k,v in self.splits.items():
            split_features = []
            split_annotations = []
            for f in v:
                if f not in self.features:
                    self.logger.error('Could not find file {} in {} split. This split will be ignored'.format(f, k))
                    split_features = []
                    split_annotations = []
                    break
                split_features.append(self.features[f])
                if self.annotations:
                    if f not in self.annotations:
                        self.logger.error('Could not find annotation {} in {} split. This split will be ignored'.format(f, k))
                        split_features = []
                        split_annotations = []
                        break
                    split_annotations.append(self.annotations[f])
            if not split_features:
                continue
            features_filename = '{}-{}_features.pkl'.format(self.name, k)
            feature_filepath = os.path.join(output_dir, features_filename)
            self.logger.info('writing {}'.format(feature_filepath))
            np.array(split_features).dump(feature_filepath)
            if split_annotations:
                annotation_filename = '{}-{}_annotations.pkl'.format(self.name, k)
                annotation_filepath = os.path.join(output_dir, annotation_filename)
                self.logger.info('writing {}'.format(annotation_filepath))
                np.array(split_annotations).dump(annotation_filepath)


if __name__ == '__main__':
    if not os.path.exists('logs'):
        os.makedirs('logs')
    logging.config.dictConfig(logging_dict)
    logger = logging.getLogger('dataset_creator')
    if len(sys.argv) != 2 or not os.path.isdir(sys.argv[1]):
        logger.error('You need to provide a dataset')
        exit()
    dataset = sys.argv[1]
    key_profiles = [
        'aarden_essen',
        'temperley',
        'krumhansl_kessler',
        'bellman_budge',
        'albrecht_shanahan1',
        'albrecht_shanahan2',
        'sapp',
        'simple_natural_minor',
        'simple_harmonic_minor',
        'simple_melodic_minor', ]
    key_transitions = [
        'ktg_exponential5',
        'ktg_exponential10',
        'ktg_exponential15', ]
    dc = DatasetCreator(dataset)
    dc.compute_features(key_profiles, key_transitions)
    dc.write()



import justkeydding.dataset
import joblib
import os

if __name__ == '__main__':
    model = 'pre-trained.joblib'
    if not os.path.exists(model):
        print('You need to train a model first')
        exit()
    data = justkeydding.dataset.Dataset('albrecht_chroma')
    X = justkeydding.dataset.feature_scaling(data.X)
    y = data.y
    clf = joblib.load(model)
    score = clf.score(X, y)
    print(score)



from justkeydding.parameters.key_profiles import (
    flush_key_profiles,
    keep_key_profiles,
    log_key_profiles_dict
)
from justkeydding.parameters.key_transitions import (
    flush_key_transitions,
    keep_key_transitions,
    log_key_transitions_dict
)
from justkeydding.optimizer import generator, evaluator, evolver
import os
import logging
import logging.handlers
import logging.config

logging_dict = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'date': {
            'format': '%(asctime)s [%(name)-15s] %(levelname)-7s: %(message)s',
        },
        'no_date': {
            'format': '[%(name)-15s] %(levelname)-7s: %(message)s',
        },
    },
    'handlers': {
        'file_handler': {
            'class': 'logging.handlers.TimedRotatingFileHandler',
            'filename': 'logs/optimizer.log',
            'when': 'H',
            'interval': 1,
            'backupCount': 360,
            'formatter': 'date',
            'encoding': 'utf8'
        },
        'console': {
            'class': 'logging.StreamHandler',
            'level': 'DEBUG',
            'formatter': 'no_date'
        }
    },
    'loggers': {
        '': {
            'handlers': ['file_handler', 'console'],
            'level': 'DEBUG',
            'propagate': True
        }
    }
}


def ga_runner(dataset, population_size, initial_key_profiles, initial_key_transitions, kp_max_range, kt_max_range, evolution_swap_threshold, just_evalute=False):
    logger.info('ga_runner() <- dataset={}, population_size={}, initial_key_profiles={}, initial_key_transitions={}, kp_max_range={}, kt_max_range={}, evolution_swap_threshold={}'.format(dataset, population_size, initial_key_profiles, initial_key_transitions, kp_max_range, kt_max_range, evolution_swap_threshold))
    # Invite all the friends to the party
    gen = generator.Generator(kp_max_range, kt_max_range)
    eva = evaluator.Evaluator(dataset)
    evo = evolver.Evolver(gen)
    # Set the initial population of key profiles
    key_profiles = initial_key_profiles
    key_profiles_length = len(key_profiles)
    if population_size < key_profiles_length:
        logger.warn('Size of initial population of key profiles greater than population size, extending population size to {}.'.format(key_profiles_length))
        population_size = key_profiles_length
    # Same for the key transitions
    key_transitions = initial_key_transitions
    key_transitions_length = len(key_transitions)
    if population_size < key_transitions_length:
        logger.warn('Size of initial population of key transitions greater than population size, extending population size to {}.'.format(key_transitions_length))
        population_size = key_transitions_length
    # Generate the remaining slots available
    # remaining_key_profiles = population_size - key_profiles_length
    # generated_key_profiles = gen.generate_key_profiles(remaining_key_profiles)
    # key_profiles.extend(generated_key_profiles)
    # remaining_key_transitions = population_size - key_transitions_length
    # generated_key_transitions = gen.generate_key_transitions(remaining_key_transitions)
    # key_transitions.extend(generated_key_transitions)
    # logger.info('Generation 0 consists of {0} ({1} user-provided, {2} generated) key profiles and {0} ({3} user-provided, {4} generated) key transitions'.format(population_size, key_profiles_length, remaining_key_profiles, key_transitions_length, remaining_key_transitions))
    # Time to find the best kp,kt pair from the initial values
    best_key_profile = ''
    best_key_transition = ''
    lowest_error = float("inf")
    scores = []
    logger.info("Finding the best (key_profile,key_transition) from the initial populations")
    kt_grading = []
    full_grading = []
    for kt in key_transitions:
        grading = eva.grade_key_profiles(key_profiles, kt)
        full_grading.extend(grading)
        lower_error, kp, _ = grading[0]
        kt_grading.append((lower_error, kt))
        if lower_error < lowest_error:
            lowest_error = lower_error
            best_key_profile = kp
            best_key_transition = kt
            scores = [x[0] for x in grading]
            key_profiles = [x[1] for x in grading]
            if lowest_error == 0:
                break
        logger.debug("({},{}) is the best pair so far".format(best_key_profile, best_key_transition))
    full_grading = sorted(full_grading, key=lambda score: score[0])
    logger.debug('full_grading:{}'.format(full_grading))
    kt_grading = sorted(kt_grading, key=lambda score: score[0])
    key_transitions = [x[1] for x in kt_grading]
    logger.debug("sorted_key_transitions:{}".format(key_transitions))
    logger.info('From the initial populations, ({},{}) is the best (key_profile, key_transition) pair'.format(best_key_profile, best_key_transition))
    logger.debug('Scores for the initial generation: {}'.format(scores))
    if just_evalute:
        logger.info('Evaluation complete!')
        exit()
    kp_generation = 1
    kt_generation = 1
    bad_generation_counter = 0
    evolution_mode = 'key_profiles'
    while lowest_error > 0:
        logger.info('Key profile generation {}, Key transition generation {} - Evolving {} - Bad generations in a row: {}'.format(kp_generation, kt_generation, evolution_mode, bad_generation_counter))
        if evolution_mode == 'key_profiles':
            new_key_profiles = evo.evolve_key_profiles(key_profiles)
            new_grading = eva.grade_key_profiles(new_key_profiles, best_key_transition)
            new_scores = [x[0] for x in new_grading]
            logger.info('Scores for this generation: {}'.format(new_scores))
            new_lowest_error, new_best_kp, _ = new_grading[0]
            logger.info('Lowest error of {} from {}'.format(new_lowest_error, new_best_kp))
            if new_lowest_error < lowest_error:
                lowest_error = new_lowest_error
                best_key_profile = new_best_kp
                scores = new_scores
                key_profiles = [x[1] for x in new_grading]
                bad_generation_counter = 0
                keep_key_profiles(key_profiles)
                log_key_profiles_dict()
            elif new_lowest_error == lowest_error and sum(new_scores) <= sum(scores):
                lowest_error = new_lowest_error
                best_key_profile = new_best_kp
                scores = new_scores
                key_profiles = [x[1] for x in new_grading]
                bad_generation_counter = 0
                keep_key_profiles(key_profiles)
                log_key_profiles_dict()
            else:
                logger.warn('Performance was worst in this generation')
                bad_generation_counter += 1
                flush_key_profiles([k for k in new_key_profiles if k not in key_profiles])
                if bad_generation_counter > evolution_swap_threshold:
                    evolution_mode = 'key_transitions'
                    bad_generation_counter = 0
                    logger.warn("Changing evolution mode to {}".format(evolution_mode))
            kp_generation += 1
        elif evolution_mode == 'key_transitions':
            new_key_transitions = evo.evolve_key_transitions(key_transitions)
            new_grading = eva.grade_key_transitions(best_key_profile, new_key_transitions)
            new_scores = [x[0] for x in new_grading]
            logger.info('Scores for this generation: {}'.format(new_scores))
            new_lowest_error, _, new_best_kt = new_grading[0]
            logger.info('Lowest error of {} from {}'.format(new_lowest_error, new_best_kt))
            if new_lowest_error < lowest_error:
                lowest_error = new_lowest_error
                best_key_transition = new_best_kt
                scores = new_scores
                key_transitions = [x[2] for x in new_grading]
                bad_generation_counter = 0
                keep_key_transitions(key_transitions)
                log_key_transitions_dict()
            elif new_lowest_error == lowest_error and sum(new_scores) <= sum(scores):
                lowest_error = new_lowest_error
                best_key_transition = new_best_kt
                scores = new_scores
                key_transitions = [x[2] for x in new_grading]
                bad_generation_counter = 0
                keep_key_transitions(key_transitions)
                log_key_transitions_dict()
            else:
                logger.warn('Performance was worst in this generation')
                bad_generation_counter += 1
                flush_key_transitions([k for k in new_key_transitions if k not in key_transitions])
                if bad_generation_counter > evolution_swap_threshold:
                    evolution_mode = 'key_profiles'
                    bad_generation_counter = 0
                    logger.warn("Changing evolution mode to {}".format(evolution_mode))
            kt_generation += 1
        logger.info('Best key_profile,key_transition pair ({}, {})'.format(best_key_profile, best_key_transition))
    logger.warn('Optimization concluded, best (key_profile, key_transition) pair is ({},{})'.format(best_key_profile, best_key_transition))


if __name__ == '__main__':
    if not os.path.exists('logs'):
        os.makedirs('logs')
    logging.config.dictConfig(logging_dict)
    logger = logging.getLogger('ga_runner')
    dataset = 'midi_dataset.txt'
    population_size = 3
    kp_max_range = 100
    kt_max_range = 256
    evolution_swap_threshold = 3
    initial_key_profiles = [
        'sapp', 
        'temperley', 
        'krumhansl_kessler', 
        'aarden_essen', 
        'albrecht_shanahan1', 
        'albrecht_shanahan2', 
        'bellman_budge', 
        'simple_natural_minor', 
        'simple_harmonic_minor', 
        'simple_melodic_minor']
    initial_key_transitions = [
        'ktg_exponential5', 
        'ktg_exponential10',
        'ktg_exponential15']
    ga_runner(dataset, population_size, initial_key_profiles, initial_key_transitions, kp_max_range, kt_max_range, evolution_swap_threshold, True)



import madmom.audio.chroma as chroma
import numpy as np
import numpy.linalg as linalg

supported_extensions = ['wav']


def _cosine_similarity(u, v):
    dot = np.dot(u, v)
    cos_sim = dot / (linalg.norm(u) * linalg.norm(v))
    return cos_sim


def parse_file(filename):
    deepchroma = chroma.DeepChromaProcessor()
    deepchroma = deepchroma(filename)
    similarity_threshold = 0.96
    threshold = 0.5
    prevframe = np.array([0.01] * 12)
    slices = []
    times = []
    for idx, frame in enumerate(deepchroma):
        cos_sim = _cosine_similarity(frame, prevframe)
        # print(np.around(frame, 1), np.around(cos_sim, 2), end='')
        if cos_sim > similarity_threshold:
            # print(' ignoring')
            continue
        slic = np.argwhere(frame > threshold)
        slices.append(slic.reshape(1, -1)[0].tolist())
        # TODO: For now, the "times" signal stored for audio is simply
        # the number of frame that is being stored. A better alternative
        # would be the time (in seconds) where this frame occurs, from
        # the beginning of the audio file
        times.append(idx)
        prevframe = frame
    return slices, times




import music21

supported_extensions = ['mxl', 'musicxml', 'xml', 'krn', 'mei']


def parse_file(filename):
    s = music21.converter.parse(filename)
    c = s.chordify()
    slices = []
    times = []
    for ev in c.flat.notes:
        if type(ev) == music21.chord.Chord:
            slic = [n.pitch.pitchClass for n in ev]
        elif type(ev) == music21.note.Note:
            slic = [ev.pitch.pitchClass]
        else:
            slic = []
        slices.append(slic)
        times.append(ev.offset)
    return slices, times


def annotate_local_keys(filename, local_keys):
    s = music21.converter.parse(filename)
    c = s.chordify()
    c.partName = 'Local keys'
    c.partAbbreviation = 'Local keys'
    s.append(c)
    events = c.flat.notes
    if len(events) != len(local_keys):
        print(len(events), len(local_keys))
        raise ValueError('Expected same slice/local keys length.')
    for idx, ev in enumerate(events):
        if type(ev) == music21.chord.Chord:
            key = local_keys[idx]
            ev._notes = [music21.note.Note(key)]
            ev.addLyric(key)
        else:
            ev.addLyric('')
    path, extension = tuple(filename.rsplit('.', 1))
    new_filename = path + '_justkeydding.' + extension
    s.write(fp=new_filename, fmt='musicxml')
